from datetime import datetime
from ftw.builder import Builder
from ftw.testbrowser import browsing
from ftw.upgrade.tests.base import JsonApiTestCase
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.factory import _DEFAULT_PROFILE
import re
import transaction


class TestPloneSiteJsonApi(JsonApiTestCase):

    @browsing
    def test_api_discovery(self, browser):
        self.api_request('GET', '')

        self.assert_json_equal(
            {'api_version': 'v1',
             'actions':
                 [{'description': ('Executes a list of profiles, each '
                                   'identified by their ID.'),
                   'name': 'execute_profiles',
                   'request_method': 'POST',
                   'required_params': ['profiles:list']},

                  {'name': 'execute_proposed_upgrades',
                   'required_params': [],
                   'description': 'Executes all proposed upgrades.',
                   'request_method': 'POST'},

                  {'name': 'execute_upgrades',
                   'required_params': ['upgrades:list'],
                   'description': 'Executes a list of upgrades, each identified by'
                   ' the upgrade ID in the form "[dest-version]@[profile ID]".',
                   'request_method': 'POST'},

                  {'name': 'get_profile',
                   'required_params': ['profileid'],
                   'description': 'Returns the profile with the id "profileid" as hash.',
                   'request_method': 'GET'},

                  {'name': 'list_profiles',
                   'required_params': [],
                   'description': 'Returns a list of all installed profiles.',
                   'request_method': 'GET'},

                  {'name': 'list_profiles_proposing_upgrades',
                   'required_params': [],
                   'description': 'Returns a list of profiles with proposed upgrade steps.'
                   ' The upgrade steps of each profile only include proposed upgrades.',
                   'request_method': 'GET'},

                  {'name': 'list_proposed_upgrades',
                   'required_params': [],
                   'description': 'Returns a list of proposed upgrades.',
                   'request_method': 'GET'},

                  {'name': 'plone_upgrade',
                   'required_params': [],
                   'description': 'Upgrade the Plone Site. This is what you would manually do in the @@plone-upgrade view.',
                   'request_method': 'POST'},

                  {'name': 'plone_upgrade_needed',
                   'required_params': [],
                   'description': 'Returns "true" when Plone needs to be upgraded.',
                   'request_method': 'GET'},

                  {'name': 'recook_resources',
                   'required_params': [],
                   'description': 'Recook CSS and JavaScript resource bundles.',
                   'request_method': 'POST'}]},

            browser.json)

    @browsing
    def test_get_profile(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1', to='2'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            self.api_request('GET', 'get_profile', dict(profileid='the.package:default'))
            self.assertEqual('application/json; charset=utf-8',
                             browser.headers.get('content-type'))

            self.assert_json_equal(
                {'id': 'the.package:default',
                 'title': 'the.package',
                 'product': 'the.package',
                 'db_version': '1',
                 'fs_version': '20110101000000',
                 'outdated_fs_version': False,
                 'upgrades': [

                        {'id': '2@the.package:default',
                         'title': '',
                         'source': '1',
                         'dest': '2',
                         'proposed': True,
                         'done': False,
                         'orphan': False,
                         'outdated_fs_version': False},

                        {'id': '20110101000000@the.package:default',
                         'title': 'Upgrade.',
                         'source': '2',
                         'dest': '20110101000000',
                         'proposed': True,
                         'done': False,
                         'orphan': False,
                         'outdated_fs_version': False},
                        ]},

                browser.json)

    @browsing
    def test_get_profile_requires_profileid(self, browser):
        with self.expect_api_error(status=400,
                                   message='Param missing',
                                   details='The param "profileid" is required'
                                   ' for this API action.'):
            self.api_request('GET', 'get_profile')

    @browsing
    def test_get_profile_requires_GET(self, browser):
        with self.expect_api_error(status=405,
                                   message='Method Not Allowed',
                                   details='Action requires GET') as cm:
            self.api_request('POST', 'get_profile', {'profileid': 'the.package:default'})
        self.assertEquals('GET', cm['headers'].get('allow'))

    @browsing
    def test_get_unkown_profile_returns_error(self, browser):
        with self.expect_api_error(status=400,
                                   message='Profile not available',
                                   details='The profile "something:default" is wrong'
                                   ' or not installed on this Plone site.'):
            self.api_request('GET', 'get_profile', {'profileid': 'something:default'})

    @browsing
    def test_cyclic_dependency_errors_are_handled(self, browser):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('foo')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:bar'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('bar')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:foo'))

        with self.package_created():
            self.install_profile('the.package:foo')
            self.install_profile('the.package:bar')

            with self.expect_api_error(status=500,
                                       message='Cyclic dependencies',
                                       details='There are cyclic Generic Setup profile'
                                       ' dependencies.'):
                self.api_request('GET', 'get_profile', {'profileid': 'the.package:foo'})

    @browsing
    def test_list_profiles(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))
        self.package.with_profile(
            Builder('genericsetup profile')
            .named('foo')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default')
            self.install_profile('the.package:foo')

            self.api_request('GET', 'list_profiles')
            self.assertEqual('application/json; charset=utf-8',
                             browser.headers.get('content-type'))

            self.assert_json_contains_profile(
                {'id': 'the.package:default',
                 'title': 'the.package',
                 'product': 'the.package',
                 'db_version': '20110101000000',
                 'fs_version': '20110101000000',
                 'outdated_fs_version': False,
                 'upgrades': [
                        {'id': '20110101000000@the.package:default',
                         'title': 'Upgrade.',
                         'source': '10000000000000',
                         'dest': '20110101000000',
                         'proposed': False,
                         'done': True,
                         'orphan': False,
                         'outdated_fs_version': False},
                        ]},
                browser.json)

            self.assert_json_contains_profile(
                {'id': 'the.package:foo',
                 'title': 'the.package',
                 'product': 'the.package',
                 'db_version': '20110101000000',
                 'fs_version': '20110101000000',
                 'outdated_fs_version': False,
                 'upgrades': [
                        {'id': '20110101000000@the.package:foo',
                         'title': 'Upgrade.',
                         'source': '10000000000000',
                         'dest': '20110101000000',
                         'proposed': False,
                         'done': True,
                         'orphan': False,
                         'outdated_fs_version': False},
                        ]},
                browser.json)

    @browsing
    def test_list_profiles_requires_authentication(self, browser):
        with self.expect_api_error(status=401,
                                   message='Unauthorized',
                                   details='Admin authorization required.'):
            self.api_request('GET', 'list_profiles', authenticate=False)

    @browsing
    def test_list_profiles_proposing_upgrades(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1', to='2'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='2')
            self.clear_recorded_upgrades('the.package:default')

            self.api_request('GET', 'list_profiles_proposing_upgrades')
            self.assertEqual('application/json; charset=utf-8',
                             browser.headers.get('content-type'))

            self.assert_json_contains_profile(
                {'id': 'the.package:default',
                 'title': 'the.package',
                 'product': 'the.package',
                 'db_version': '2',
                 'fs_version': '20110101000000',
                 'outdated_fs_version': False,
                 'upgrades': [
                        {'id': '20110101000000@the.package:default',
                         'title': 'Upgrade.',
                         'source': '2',
                         'dest': '20110101000000',
                         'proposed': True,
                         'done': False,
                         'orphan': False,
                         'outdated_fs_version': False},
                        ]},
                browser.json)

    @browsing
    def test_list_proposed_upgrades_when_empty(self, browser):
        self.api_request('GET', 'list_proposed_upgrades')
        self.assertEqual('application/json; charset=utf-8',
                         browser.headers.get('content-type'))

    @browsing
    def test_list_proposed_upgrdes(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1', to='2')))

        self.package.with_profile(
            Builder('genericsetup profile')
            .named('foo')
            .with_upgrade(Builder('plone upgrade step').upgrading('2', to='3')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.install_profile('the.package:foo', version='1')

            self.api_request('GET', 'list_proposed_upgrades')
            self.assertEqual('application/json; charset=utf-8',
                             browser.headers.get('content-type'))

            self.assert_json_contains(
                {'id': '2@the.package:default',
                 'title': '',
                 'source': '1',
                 'dest': '2',
                 'proposed': True,
                 'done': False,
                 'orphan': False,
                 'outdated_fs_version': False},
                browser.json)

            self.assert_json_contains(
                {'id': '3@the.package:foo',
                 'title': '',
                 'source': '2',
                 'dest': '3',
                 'proposed': True,
                 'done': False,
                 'orphan': False,
                 'outdated_fs_version': False},
                browser.json)

    @browsing
    def test_execute_upgrades_installs_upgrades_in_gatherer_order(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The first upgrade step'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))
                          .named('The second upgrade step')))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')
            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertFalse(self.is_installed('the.package:default', datetime(2012, 2, 2)))

            self.api_request('POST', 'execute_upgrades', (
                    ('upgrades:list', '20120202000000@the.package:default'),
                    ('upgrades:list', '20110101000000@the.package:default')))
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.assertTrue(self.is_installed('the.package:default', datetime(2012, 2, 2)))
            self.assertEqual(
                ['UPGRADE STEP the.package:default: The first upgrade step.',
                 'UPGRADE STEP the.package:default: The second upgrade step.'],
                re.findall(r'UPGRADE STEP.*', browser.contents))

            self.assertIn('Result: SUCCESS', browser.contents)

    @browsing
    def test_execute_upgrades_requires_upgrades_param(self, browser):
        with self.expect_api_error(status=400,
                                   message='Param missing',
                                   details='The param "upgrades:list" is required for'
                                   ' this API action.'):
            self.api_request('POST', 'execute_upgrades')

    @browsing
    def test_execute_upgrades_validates_upgrade_ids(self, browser):
        with self.expect_api_error(status=400,
                                   message='Upgrade not found',
                                   details='The upgrade "foo@bar:default" is unkown.'):
            self.api_request('POST', 'execute_upgrades', {'upgrades:list': 'foo@bar:default'})

    @browsing
    def test_execute_upgrades_requires_POST(self, browser):
        with self.expect_api_error(status=405,
                                   message='Method Not Allowed',
                                   details='Action requires POST') as cm:
            self.api_request('GET', 'execute_upgrades', {'upgrades:list': 'foo@bar:default'})
        self.assertEquals('POST', cm['headers'].get('allow'))

    @browsing
    def test_execute_upgrades_not_allowed_when_plone_outdated(self, browser):
        portal_migration = getToolByName(self.layer['portal'], 'portal_migration')
        portal_migration.setInstanceVersion('1.0.0')
        transaction.commit()

        with self.expect_api_error(status=400,
                                   message='Plone site outdated',
                                   details='The Plone site is outdated and needs to'
                                   ' be upgraded first using the regular Plone'
                                   ' upgrading tools.'):
            self.api_request('POST', 'execute_upgrades', {'upgrades:list': 'foo@bar:default'})

    @browsing
    def test_execute_proposed_upgrades(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.install_profile('the.package:default', version='2')
            self.clear_recorded_upgrades('the.package:default')

            self.assertFalse(self.is_installed('the.package:default', datetime(2011, 1, 1)))
            self.api_request('POST', 'execute_proposed_upgrades')
            self.assertTrue(self.is_installed('the.package:default', datetime(2011, 1, 1)))

            self.assertIn('UPGRADE STEP the.package:default: The upgrade.',
                          browser.contents)
            self.assertIn('Result: SUCCESS', browser.contents)

    @browsing
    def test_executing_upgrades_with_failure_results_in_error_result(self, browser):
        def failing_upgrade(setup_context):
            raise KeyError('foo')

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1', to='2')
                          .calling(failing_upgrade)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.api_request('POST', 'execute_proposed_upgrades')
            self.assertIn('Result: FAILURE', browser.contents)

    @browsing
    def test_recook_resources(self, browser):
        with self.assert_resources_recooked():
            self.api_request('POST', 'recook_resources')
            self.assertEqual('OK', browser.json)

    @browsing
    def test_execute_profiles_standard(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.api_request('POST', 'execute_profiles',
                             {'profiles:list': ['the.package:default']})
            self.assertEqual(self.portal_setup.getLastVersionForProfile(
                'the.package:default'), ('20110101000000', ))
            self.assertTrue(self.is_installed(
                'the.package:default', datetime(2011, 1, 1)))
            self.assertIn('Done installing profile the.package:default.',
                          browser.contents)
            self.assertIn('Result: SUCCESS', browser.contents)

    @browsing
    def test_execute_profiles_already_installed(self, browser):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))
                          .named('The upgrade')))

        with self.package_created():
            self.portal_setup.runAllImportStepsFromProfile(
                'profile-the.package:default')
            transaction.commit()
            self.api_request('POST', 'execute_profiles',
                             {'profiles:list': ['the.package:default']})
            self.assertIn(
                'Ignoring already installed profile the.package:default.',
                browser.contents)
            self.assertIn('Result: SUCCESS', browser.contents)

    @browsing
    def test_execute_profiles_not_found(self, browser):
        with self.expect_api_error(
                status=400,
                message='Profile not found',
                details='The profile "the.package:default" is unknown.'):
            self.api_request('POST', 'execute_profiles',
                             {'profiles:list': ['the.package:default']})

    @browsing
    def test_plone_upgrade_needed(self, browser):
        self.api_request('GET', 'plone_upgrade_needed')
        self.assertEquals('false', browser.contents.strip())

        self.portal_setup.setLastVersionForProfile(_DEFAULT_PROFILE, '4')
        transaction.commit()
        self.api_request('GET', 'plone_upgrade_needed')
        self.assertEquals('true', browser.contents.strip())



from AccessControl.SecurityInfo import ClassSecurityInformation
from Products.CMFCore.utils import getToolByName
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.utils import format_duration
from zope.component import getAdapter
from zope.publisher.browser import BrowserView
import logging
import time
import traceback


LOG = logging.getLogger('ftw.upgrade')


class ResponseLogger(object):

    security = ClassSecurityInformation()

    def __init__(self, response, annotate_result=False):
        self.response = response
        self.handler = None
        self.formatter = None
        self.annotate_result = annotate_result

    def __enter__(self):
        self.handler = logging.StreamHandler(self)
        self.formatter = logging.root.handlers[-1].formatter
        self.handler.setFormatter(self.formatter)
        logging.root.addHandler(self.handler)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            LOG.error('FAILED')
            traceback.print_exception(exc_type, exc_value, tb, None, self)
            if self.annotate_result:
                self.write('Result: FAILURE\n')

        elif self.annotate_result:
            self.write('Result: SUCCESS\n')

        logging.root.removeHandler(self.handler)

        # Plone testing does not collect data written to the response stream
        # but only data set directly as body.
        # Since we want to test the response body, we need to re-set the
        # stream data as body for testing..
        if self.response.__class__.__name__ == 'TestResponse':
            self.response.setBody(self.response.stdout.getvalue())

    security.declarePrivate('write')
    def write(self, line):
        if isinstance(line, unicode):
            line = line.encode('utf8')

        self.response.write(line)
        self.response.flush()

    security.declarePrivate('writelines')
    def writelines(self, lines):
        for line in lines:
            self.write(line)


class ManageUpgrades(BrowserView):

    security = ClassSecurityInformation()

    def __init__(self, *args, **kwargs):
        super(ManageUpgrades, self).__init__(*args, **kwargs)
        self.cyclic_dependencies = False

    def __call__(self):
        if self.request.get('submitted', False):
            assert not self.plone_needs_upgrading(), \
                'Plone is outdated. Upgrading add-ons is disabled.'

            if self.request.get('ajax', False):
                return self.install_with_ajax_stream()

            else:
                self.install()

        return super(ManageUpgrades, self).__call__(self)

    security.declarePrivate('install')
    def install(self):
        """Installs the selected upgrades.
        """
        start = time.time()

        gstool = getToolByName(self.context, 'portal_setup')
        executioner = getAdapter(gstool, IExecutioner)
        data = self._get_upgrades_to_install()
        executioner.install(data)

        logging.getLogger('ftw.upgrade').info('FINISHED')

        logging.getLogger('ftw.upgrade').info(
            'Duration for all selected upgrade steps: %s' % (
                format_duration(time.time() - start)))

    security.declarePrivate('install_with_ajax_stream')
    def install_with_ajax_stream(self):
        """Installs the selected upgrades and streams the log into
        the HTTP response.
        """
        response = self.request.RESPONSE
        response.setHeader('Content-Type', 'text/html')
        response.setHeader('Transfer-Encoding', 'chunked')
        response.write('<html>')
        response.write('<body>')
        response.write('  ' * getattr(response, 'http_chunk_size', 100))
        response.write('<pre>')

        with ResponseLogger(self.request.RESPONSE):
            self.install()

        response.write('</pre>')
        response.write('</body>')
        response.write('</html>')

    security.declarePrivate('get_data')
    def get_data(self):
        gstool = getToolByName(self.context, 'portal_setup')
        gatherer = getAdapter(gstool, IUpgradeInformationGatherer)
        try:
            return gatherer.get_profiles()
        except CyclicDependencies, exc:
            self.cyclic_dependencies = exc.cyclic_dependencies
            return []

    security.declarePrivate('plone_needs_upgrading')
    def plone_needs_upgrading(self):
        portal_migration = getToolByName(self.context, 'portal_migration')
        return portal_migration.needUpgrading()

    security.declarePrivate('_get_upgrades_to_install')
    def _get_upgrades_to_install(self):
        """Returns a dict where the key is a profileid and the value
        is a list of upgrade ids.
        """

        data = {}
        for item in self.request.get('upgrade', []):
            item = dict(item)
            profileid = item['profileid']
            del item['profileid']

            if item:
                data[profileid] = item.keys()

        upgrades = []

        for profile in self.get_data():
            if profile.get('id') not in data:
                continue

            profile_data = data[profile.get('id')]
            if not profile.get('upgrades', []):
                continue

            profile_upgrades = []
            upgrades.append((profile.get('id'), profile_upgrades))

            for upgrade in profile.get('upgrades', []):
                if upgrade.get('id') in profile_data:
                    profile_upgrades.append(upgrade.get('id'))

        return upgrades


class ManageUpgradesPlain(ManageUpgrades):

    def __getitem__(self, key):
        return self.index.macros[key]



from ftw.upgrade.utils import SavepointIterator
from unittest2 import TestCase
import transaction


class TestSavepointIterator(TestCase):

    def setUp(self):
        self.iterable = [1, 2, 3, 4, 5]
        self.txn = transaction.get()

    def tearDown(self):
        self.txn.abort()

    def test_creates_savepoints(self):
        self.assertEquals(
            0, self.txn._savepoint_index,
            'A new transaction should not have any savepoints yet')

        iterator = SavepointIterator.build(self.iterable, threshold=5)

        # Consume entire iterator
        result = list(iterator)

        self.assertEquals(
            self.iterable, result,
            'Iterator should yield every item of `iterable`')

        self.assertEquals(
            1, self.txn._savepoint_index,
            'One savepoint should have been created')

    def test_doesnt_create_savepoints_with_threshold_0(self):
        self.assertEquals(
            0, self.txn._savepoint_index,
            'A new transaction should not have any savepoints yet')

        iterator = SavepointIterator.build(self.iterable, threshold=0)

        # Consume entire iterator
        result = list(iterator)

        self.assertEquals(
            self.iterable, result,
            'Iterator should yield every item of `iterable`')

        self.assertEquals(
            0, self.txn._savepoint_index,
            'threshold=0 should never create any savepoints')

    def test_instanciating_iterator_with_nonzero_threshold_raises(self):
        with self.assertRaises(ValueError):
            SavepointIterator(self.iterable, threshold=0)

        with self.assertRaises(ValueError):
            SavepointIterator(self.iterable, threshold=None)



from AccessControl import getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import SpecialUsers
from Acquisition import aq_inner
from Acquisition import aq_parent
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.exceptions import UpgradeNotFound
from ftw.upgrade.jsonapi.exceptions import AbortTransactionWithStreamedResponse
from ftw.upgrade.jsonapi.exceptions import APIError
from ftw.upgrade.jsonapi.exceptions import CyclicDependenciesWrapper
from ftw.upgrade.jsonapi.exceptions import MethodNotAllowed
from ftw.upgrade.jsonapi.exceptions import MissingParam
from ftw.upgrade.jsonapi.exceptions import UnauthorizedWrapper
from ftw.upgrade.jsonapi.exceptions import UpgradeNotFoundWrapper
from ftw.upgrade.utils import get_tempfile_authentication_directory
from OFS.interfaces import IApplication
from zExceptions import Unauthorized
from zope.security import checkPermission
import inspect
import json
import os
import re
import stat
import transaction


class ErrorHandling(object):
    """Context manager for handling API errors and responding as JSON.
    """

    exception_wrappers = {
        CyclicDependencies: CyclicDependenciesWrapper,
        UpgradeNotFound: UpgradeNotFoundWrapper,
        Unauthorized: UnauthorizedWrapper}

    def __init__(self, response):
        self.response = response

    def __enter__(self):
        return self

    def __exit__(self, _type, exc, _traceback):
        if isinstance(exc, AbortTransactionWithStreamedResponse):
            if isinstance(self.wrap_exception(exc.original_exception),
                          APIError):
                exc = exc.original_exception
            else:
                transaction.abort()
                return True

        exc = self.wrap_exception(exc)
        if not isinstance(exc, APIError):
            return

        self.response.setStatus(exc.response_code, exc.message)
        self.response.setHeader('Content-Type',
                                'application/json; charset=utf-8')
        exc.process_error(self.response)
        self.response.setBody(
            json.dumps(['ERROR', exc.message, exc.details]) + '\n')
        self.response.flush()
        return True

    def wrap_exception(self, original_exception):
        for original_type, wrapper_type in self.exception_wrappers.items():
            if isinstance(original_exception, original_type):
                return wrapper_type(original_exception)
        return original_exception


def action(method, rename_params={}):
    """Decorats an API action.
    The action is protected to only respond to one HTTP method
    and protects the action by the cmf.ManagePortal permission.
    Known API errors are written as JSON to the respond.
    """
    def wrap_action(func):
        def action_wrapper(self):
            with ErrorHandling(self.request.RESPONSE):
                if self.request.method != method:
                    raise MethodNotAllowed(method)

                perform_tempfile_authentication(self.context, self.request)

                if not checkPermission('cmf.ManagePortal', self.context):
                    raise Unauthorized()

                params = extract_action_params(
                    func, self.request, rename_params)
                return func(self, **params)

        action_wrapper.__doc__ = func.__doc__
        action_wrapper.__name__ = func.__name__
        action_wrapper.action_info = {
            'method': method,
            'rename_params': rename_params,
            'name': func.__name__,
            'doc': func.__doc__,
            'argspec': inspect.getargspec(func)}
        return action_wrapper
    return wrap_action


def jsonify(func):
    """Action decorator for converting response data to JSON.
    """

    def json_wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        response = self.request.RESPONSE
        if 'json' in (response.getHeader('Content-Type') or ''):
            # already converted to json, e.g. on error.
            return result

        response.setHeader('Content-Type', 'application/json; charset=utf-8')
        return json.dumps(result, indent=4, encoding='utf-8') + '\n'

    json_wrapper.__doc__ = func.__doc__
    json_wrapper.__name__ = func.__name__
    json_wrapper.action_info = getattr(func, 'action_info', None)
    return json_wrapper


def extract_action_params(func, request, rename_params=None):
    rename_params = rename_params or {}
    form = request.form
    argspec = inspect.getargspec(func)
    required_params = get_required_args(argspec)

    for arg_name in required_params:
        if not form.get(arg_name, None):
            raise MissingParam(rename_params.get(arg_name, arg_name))

    return dict([(name, form[name]) for name in form if name in argspec.args])


def get_action_discovery_information(view):
    result = []
    for name in sorted(dir(view)):
        if name == '__call__':
            continue

        func = getattr(view, name, None)
        if not func:
            continue

        action_info = getattr(func, 'action_info', None)
        if not action_info:
            continue

        argspec = action_info['argspec']
        required_params = sorted(get_required_args(argspec))
        rename_params = action_info['rename_params']
        required_params = [rename_params.get(name, name)
                           for name in required_params]

        result.append({
                'name': action_info['name'],
                'description': re.sub(r'\s+', ' ', action_info['doc']).strip(),
                'required_params': required_params,
                'request_method': action_info['method'].upper()})

    return result


def get_required_args(argspec):
    if not argspec.defaults:
        return argspec.args[1:]
    else:
        return argspec.args[1:-len(argspec.defaults)]


def perform_tempfile_authentication(context, request):
    """When the "x-ftw.upgrade-tempfile-auth" header is set, authentication is
    initialized based on a tempfile value for verifying that the client and the
    server is on the same machine with the same user.

    When necessary, a system-upgrade user is created with Manager role and
    the security is set to this user.
    """
    if getSecurityManager().getUser() != SpecialUsers.nobody:
        return

    auth_value = request.getHeader('x-ftw.upgrade-tempfile-auth')
    if not auth_value:
        return

    validate_tempfile_authentication_header_value(auth_value)
    user = get_system_upgrade_user(context)
    newSecurityManager(request, user)
    transaction.get().setUser(user.getId(), '')


def validate_tempfile_authentication_header_value(header_value):
    if not re.match('^tmp\w{6}:\w{64}', header_value):
        raise ValueError(
            'tempfile auth: invalid x-ftw.upgrade-tempfile-auth header value.')

    filename, authhash = header_value.split(':')
    directory = get_tempfile_authentication_directory(os.getcwd())
    filepath = directory.joinpath(filename)

    if not filepath.isfile():
        raise ValueError('tempfile auth: tempfile does not exist.')

    if stat.S_IMODE(filepath.stat().st_mode) != 0600:
        raise ValueError('tempfile auth: tempfile has invalid mode.')

    if filepath.stat().st_uid != os.getuid():
        raise ValueError('tempfile auth: tempfile has invalid owner.')

    if filepath.getsize() != 64:
        raise ValueError('tempfile auth: tempfile size is invalid.')

    with open(filepath, 'r') as authfile:
        if authfile.read() != authhash:
            raise ValueError('tempfile auth: authentication failed.')


def get_system_upgrade_user(context):
    while not IApplication.providedBy(context):
        context = aq_parent(aq_inner(context))

    acl_users = context.acl_users
    if not acl_users.getUserById('system-upgrade'):
        acl_users.userFolderAddUser(
            'system-upgrade', os.urandom(16).encode('hex'), ['Manager'], None)
    return acl_users.getUserById('system-upgrade')



from AccessControl.SecurityInfo import ClassSecurityInformation
from datetime import datetime
from ftw.upgrade.exceptions import UpgradeNotFound
from ftw.upgrade.interfaces import IRecordableHandler
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from ftw.upgrade.utils import get_sorted_profile_ids
from operator import itemgetter
from Products.CMFCore.utils import getToolByName
from Products.GenericSetup.interfaces import ISetupTool
from Products.GenericSetup.upgrade import normalize_version
from Products.GenericSetup.upgrade import UpgradeStep
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.deprecation import deprecated
from zope.interface import implements


def flatten_upgrades(upgrades):
    """Flattens the data structure of a list of upgrades: removes grouping.
    The result is an iterable with dicts containg upgrade information.
    """

    for item in upgrades:
        if isinstance(item, (list, tuple)):
            for subitem in flatten_upgrades(item):
                yield subitem

        else:
            yield item


def flag_profiles_with_outdated_fs_version(upgrades):
    """Flags profiles that contain an upgrade that leads to a destination
    version that is higher than that profile's current filesystem version.

    This usually means someone wrote an upgrade step, but forgot to update the
    version in metadata.xml of the corresponding profile. The upgrade step in
    question will also be flagged.
    """

    for profile in upgrades:
        profile['outdated_fs_version'] = False
        fs_version = normalize_version(profile['version'])

        for upgrade in profile['upgrades']:
            dest_version = normalize_version(upgrade['sdest'])
            upgrade['outdated_fs_version'] = dest_version > fs_version
            if upgrade['outdated_fs_version']:
                profile['outdated_fs_version'] = True

    return upgrades


def extend_auto_upgrades_with_human_formatted_date_version(profiles):
    """Adds a 'fsource' and / or 'fdest' key to each upgrade dist where the
    corresponding version is a 14 digit timestamp with the timestamp in a
    human readable format.
    """
    to_human_readable = lambda datestr: datetime.strptime(datestr, '%Y%m%d%H%M%S') \
        .strftime('%Y/%m/%d %H:%M')

    for profile in profiles:
        if len(profile.get('db_version', '')) == 14:
            try:
                profile['formatted_db_version'] = to_human_readable(
                    profile['db_version'])
            except ValueError:
                pass

        if len(profile.get('version', '')) == 14:
            try:
                profile['formatted_version'] = to_human_readable(
                    profile['version'])
            except ValueError:
                pass

        for upgrade in profile['upgrades']:
            if len(upgrade['ssource']) == 14:
                try:
                    upgrade['fsource'] = to_human_readable(upgrade['ssource'])
                except ValueError:
                    pass

            if len(upgrade['sdest']) == 14:
                try:
                    upgrade['fdest'] = to_human_readable(upgrade['sdest'])
                except ValueError:
                    pass

    return profiles


class UpgradeInformationGatherer(object):
    implements(IUpgradeInformationGatherer)
    adapts(ISetupTool)

    security = ClassSecurityInformation()

    def __init__(self, portal_setup):
        self.portal_setup = portal_setup
        self.portal = getToolByName(
            portal_setup, 'portal_url').getPortalObject()
        self.cyclic_dependencies = False

    security.declarePrivate('get_profiles')
    def get_profiles(self, proposed_only=False):
        profiles = self._sort_profiles_by_dependencies(
            self._get_profiles(proposed_only=proposed_only))
        profiles = flag_profiles_with_outdated_fs_version(profiles)
        profiles = extend_auto_upgrades_with_human_formatted_date_version(
            profiles)
        return profiles

    security.declarePrivate('get_upgrades')
    get_upgrades = deprecated(get_profiles,
                              'get_upgrades was renamed to get_profiles')

    security.declarePrivate('get_upgrades_by_api_ids')
    def get_upgrades_by_api_ids(self, *api_ids):
        upgrades = filter(lambda upgrade: upgrade['api_id'] in api_ids,
                          reduce(list.__add__,
                                 map(itemgetter('upgrades'),
                                     self.get_profiles())))
        missing_api_ids = (set(api_ids)
                           - set(map(itemgetter('api_id'), upgrades)))
        if missing_api_ids:
            raise UpgradeNotFound(tuple(missing_api_ids)[0])
        return upgrades

    security.declarePrivate('_get_profiles')
    def _get_profiles(self, proposed_only=False):
        for profileid in self.portal_setup.listProfilesWithUpgrades():
            if not self._is_profile_installed(profileid):
                continue

            data = self._get_profile_data(
                profileid, proposed_only=proposed_only)
            if len(data['upgrades']) == 0:
                continue

            if profileid == 'Products.CMFPlone:plone':
                # Plone has its own migration mechanism.
                # We do not support upgrading plone.
                continue

            yield data

    security.declarePrivate('_get_profile_data')
    def _get_profile_data(self, profileid, proposed_only=False):
        db_version = self.portal_setup.getLastVersionForProfile(profileid)
        if isinstance(db_version, (tuple, list)):
            db_version = '.'.join(db_version)

        data = {
            'upgrades': self._get_profile_upgrades(
                profileid, proposed_only=proposed_only),
            'db_version': db_version}

        try:
            profile_info = self.portal_setup.getProfileInfo(profileid).copy()
            if 'for' in profile_info:
                del profile_info['for']
            data.update(profile_info)

        except KeyError, exc:
            if exc.args and exc.args[0] == profileid:
                # package was removed - profile is no longer available.
                return {'upgrades': []}

            else:
                raise

        return data

    security.declarePrivate('_get_profile_upgrades')
    def _get_profile_upgrades(self, profileid, proposed_only=False):
        proposed_ids = set()
        upgrades = []

        proposed_upgrades = list(flatten_upgrades(
                self.portal_setup.listUpgrades(profileid)))
        all_upgrades = list(flatten_upgrades(
                self.portal_setup.listUpgrades(profileid, show_old=True)))

        for upgrade in proposed_upgrades:
            proposed_ids.add(upgrade['id'])

        for upgrade in all_upgrades:
            upgrade = upgrade.copy()
            if upgrade['id'] not in proposed_ids:
                upgrade['proposed'] = False
                upgrade['done'] = True

            upgrade['orphan'] = self._is_orphan(profileid, upgrade)
            if upgrade['orphan']:
                upgrade['proposed'] = True
                upgrade['done'] = False

            if 'step' in upgrade:
                del upgrade['step']

            upgrade['profile'] = profileid
            upgrade['api_id'] = '@'.join((upgrade['sdest'], profileid))

            if proposed_only and not upgrade['proposed']:
                continue

            upgrades.append(upgrade)

        return upgrades

    security.declarePrivate('_is_profile_installed')
    def _is_profile_installed(self, profileid):
        quickinstaller = getToolByName(self.portal_setup,
                                       'portal_quickinstaller')
        try:
            profileinfo = self.portal_setup.getProfileInfo(profileid)
        except KeyError:
            return False

        product = profileinfo['product']
        if quickinstaller.isProductInstallable(product) and \
                not quickinstaller.isProductInstalled(product):
            return False

        version = self.portal_setup.getLastVersionForProfile(profileid)
        return version != 'unknown'

    security.declarePrivate('_sort_profiles_by_dependencies')
    def _sort_profiles_by_dependencies(self, profiles):
        """Sort the profiles so that the profiles are listed after its
        dependencies since it is safer to first install dependencies.
        """

        sorted_profile_ids = get_sorted_profile_ids(self.portal_setup)
        return sorted(profiles,
                      key=lambda p: sorted_profile_ids.index(p.get('id')))

    security.declarePrivate('_is_orphan')
    def _is_orphan(self, profile, upgrade_step_info):
        if upgrade_step_info['proposed']:
            return False
        if not self._is_recordeable(upgrade_step_info):
            return False
        recorder = getMultiAdapter((self.portal, profile),
                                   IUpgradeStepRecorder)
        return not recorder.is_installed(upgrade_step_info['sdest'])

    security.declarePrivate('_is_recordeable')
    def _is_recordeable(self, upgrade_step_info):
        if not isinstance(upgrade_step_info['step'], UpgradeStep):
            return False
        handler = upgrade_step_info['step'].handler
        return IRecordableHandler.providedBy(handler)



from ftw.upgrade.tests.base import CommandTestCase


class TestHelpCommand(CommandTestCase):

    def test_help(self):
        exitcode, output = self.upgrade_script('help')
        self.assertTrue(output)

    def test_subcommand_help(self):
        exitcode, output = self.upgrade_script('help create')
        self.assertTrue(output)



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.exceptions import UpgradeNotFound
from ftw.upgrade.gatherer import extend_auto_upgrades_with_human_formatted_date_version
from ftw.upgrade.gatherer import UpgradeInformationGatherer
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.tests.base import UpgradeTestCase
from unittest2 import TestCase
from zope.component import queryAdapter
from zope.interface.verify import verifyClass


class TestUpgradeInformationGatherer(UpgradeTestCase):

    def test_implements_interface(self):
        verifyClass(IUpgradeInformationGatherer, UpgradeInformationGatherer)

    def test_not_yet_installed_upgrades_are_marked_as_proposed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1001', to='1002')))

        with self.package_created():
            self.install_profile('the.package:default', '1001')
            self.assert_gathered_upgrades({
                    'the.package:default': {'proposed': ['1002'],
                                            'done': ['1001']}})

    def test_filtering_proposed_upgrades(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1001', to='1002')))

        with self.package_created():
            self.install_profile('the.package:default', '1001')
            self.assert_gathered_upgrades(
                {'the.package:default': {'proposed': ['1002'],
                                         'done': []}},
                proposed_only=True)

    def test_profile_with_outdated_fs_version_is_flagged(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('outdated')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1111', '2222'))
                                  .with_fs_version('2222')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('2222', '3333')))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('correct')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1111', '2222'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('2222', '3333'))
                                  .with_fs_version('3333'))

        with self.package_created():
            self.install_profile('the.package:outdated', '2222')
            self.install_profile('the.package:correct', '2222')
            self.assert_outdated_profiles(['the.package:outdated'],
                                          ignore=[u'plone.app.jquerytools:default',
                                                  u'plone.app.jquery:default'])
            self.assert_gathered_upgrades({
                    'the.package:outdated': {'proposed': ['3333'],
                                            'done': ['2222'],
                                            'outdated_fs_version': ['3333']},
                    'the.package:correct': {'proposed': ['3333'],
                                            'done': ['2222'],
                                            'outdated_fs_version': []}})

    def test_profile_with_no_upgrades_is_not_listed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('without_upgrades')
                                  .with_fs_version('1'))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('with_upgrades')
                                  .with_fs_version('2')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1', to='2')))

        with self.package_created():
            self.install_profile('the.package:without_upgrades')
            self.install_profile('the.package:with_upgrades')
            self.assertIn('the.package:with_upgrades', self.get_listed_profiles(),
                          'Profiles with upgrades should be listed.')
            self.assertNotIn('the.package:without_upgrades', self.get_listed_profiles(),
                          'Profiles without upgrades should not be listed.')

    def test_profile_only_listed_when_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1', to='2')))

        with self.package_created():
            self.assertNotIn('the.package:default', self.get_listed_profiles(),
                             'Not yet installed profiles should not be listed.')

            self.install_profile('the.package:default')
            self.assertIn('the.package:default', self.get_listed_profiles(),
                          'Installed profiles should be listed.')

            self.portal_quickinstaller.uninstallProducts(['the.package'])
            self.assertNotIn('the.package:default', self.get_listed_profiles(),
                             'Packages uninstalled by quickinstaller should not be listed.')

    def test_plone_profile_is_removed(self):
        self.assertNotIn('Products.CMFPlone:plone',
                         self.get_listed_profiles(filter_package=None),
                         'Products.CMFPlone should never be listed.')

    def test_profiles_are_ordered_by_dependencies(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('one')
                                  .with_dependencies('the.package:two',
                                                     'the.package:four')
                                  .with_upgrade(self.default_upgrade()))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('two')
                                  .with_dependencies('the.package:three')
                                  .with_upgrade(self.default_upgrade()))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('three')
                                  .with_dependencies('the.package:four'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('four')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:one')
            self.install_profile('the.package:two')
            self.install_profile('the.package:three')
            self.install_profile('the.package:four')

            self.assertEquals(
                ['the.package:four',
                 'the.package:two',
                 'the.package:one'],
                self.get_listed_profiles())

    def test_cyclic_dependencies_raises_an_exception(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('foo')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:bar'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('bar')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:foo'))

        with self.package_created():
            self.install_profile('the.package:foo')
            self.install_profile('the.package:bar')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            with self.assertRaises(CyclicDependencies) as cm:
                gatherer.get_profiles()

            self.assertIn(('the.package:bar', 'the.package:foo'),
                          cm.exception.dependencies)

    def test_no_longer_existing_profiles_are_silently_removed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:removed'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('removed')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertEquals(['the.package:removed', 'the.package:default'],
                              self.get_listed_profiles())

            from Products.GenericSetup import profile_registry
            profile_registry.unregisterProfile('removed', 'the.package')

            self.assertEquals(['the.package:default'],
                              self.get_listed_profiles())

    def test_not_installable_products_are_not_checked_if_uninstalled(self):
        # Some profiles are not associated with a product.
        # Since we want those profiles to appear in manage-upgrades, we dont verify in the
        # quickinstaller that the product is installed, since it is always not installed because
        # there is no known product.
        # - Products.CMFEditions:CMFEditions
        # - Products.TinyMCE:TinyMCE
        # - plone.app.discussion:default
        # - plone.formwidget.autocomplete:default (depends on Plone version)

        profiles = self.get_listed_profiles(filter_package=None)
        self.assertIn('Products.CMFEditions:CMFEditions', profiles)
        self.assertIn('Products.TinyMCE:TinyMCE', profiles)
        self.assertIn('plone.app.discussion:default', profiles)

    def test_profile_infos(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')

            self.assertDictContainsSubset(
                {'product': 'the.package',
                 'id': u'the.package:default',
                 'db_version': u'1000',
                 'version': u'1001',
                 'outdated_fs_version': False,
                 'title': u'the.package',
                 'description': u'',
                 'type': 2,
                 },

                self.get_profiles_by_ids()['the.package:default'])

    def test_upgrade_infos(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled('Add action')
                          .with_description('Some details...')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            profile_info = self.get_profiles_by_ids()['the.package:default']
            upgrade_info = profile_info['upgrades'][0]

            self.assertDictContainsSubset(
                {'title': u'Add action',
                 'description': u'Some details...',
                 'profile': 'the.package:default',
                 'api_id': '1001@the.package:default',
                 'ssource': '1000',
                 'source': ('1000',),
                 'sdest': '1001',
                 'dest': ('1001',),
                 'done': False,
                 'proposed': True,
                 'orphan': False,
                 'outdated_fs_version': False,
                 'haspath': ('1001',)},
                upgrade_info)

    def test_orphane_upgrades_are_marked(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2013, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='20120101000000')
            self.record_installed_upgrades('the.package:default', '20120101000000')
            self.assert_gathered_upgrades({
                    'the.package:default': {
                        'done': ['1001', '20120101000000'],
                        'proposed': ['20110101000000', '20130101000000'],
                        'orphan': ['20110101000000']}})

    def test_human_readable_timestamp_versions_are_added_to_profiles(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 12, 30, 17, 45, 59))))

        with self.package_created():
            self.install_profile('the.package:default')

            self.assertDictContainsSubset(
                {'id': u'the.package:default',
                 'db_version': u'20111230174559',
                 'formatted_db_version': u'2011/12/30 17:45',
                 'version': u'20111230174559',
                 'formatted_version': u'2011/12/30 17:45'},
                self.get_profiles_by_ids()['the.package:default'])

    def test_human_readable_timestamp_versions_are_added_to_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to('20141117093040'))
            .with_upgrade(Builder('ftw upgrade step').to('20141230104550')))

        with self.package_created():
            self.install_profile('the.package:default')
            profile_info = self.get_profiles_by_ids()['the.package:default']
            upgrade_info = profile_info['upgrades'][1]

            self.assertDictContainsSubset(
                {'ssource': '20141117093040',
                 'fsource': '2014/11/17 09:30',
                 'sdest': '20141230104550',
                 'fdest': '2014/12/30 10:45'},
                upgrade_info)

    def test_get_upgrades_by_api_ids(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            upgrade_info, = gatherer.get_upgrades_by_api_ids(
                '20110101000000@the.package:default')
            self.assertDictContainsSubset(
                {'api_id': u'20110101000000@the.package:default',
                 'sdest': u'20110101000000'},
                upgrade_info)

    def test_get_upgrades_by_api_ids_orders_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            self.assertEquals(
                gatherer.get_upgrades_by_api_ids('20110101000000@the.package:default',
                                                 '20120202000000@the.package:default'),
                gatherer.get_upgrades_by_api_ids('20120202000000@the.package:default',
                                                 '20110101000000@the.package:default'))

    def test_get_upgrades_by_api_ids_raises_upgrade_not_found(self):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        with self.assertRaises(UpgradeNotFound) as cm:
            gatherer.get_upgrades_by_api_ids('foo@bar:default')
        self.assertEquals('foo@bar:default', cm.exception.api_id)
        self.assertEquals('The upgrade "foo@bar:default" could not be found.',
                          str(cm.exception))

    def get_listed_profiles(self, filter_package='the.package'):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles()
        profiles = [profile['id'] for profile in result]
        if filter_package:
            profiles = filter(lambda profile: profile.startswith(filter_package),
                              profiles)
        return profiles

    def get_profiles_by_ids(self):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles()
        return dict([(profile['id'], profile) for profile in result])

    def assert_gathered_upgrades(self, expected, *args, **kwargs):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles(*args, **kwargs)
        got = {}
        for profile in result:
            if profile['id'] not in expected:
                continue

            got_profile = dict((key, []) for key in expected[profile['id']].keys())
            got[profile['id']] = got_profile

            for upgrade in profile['upgrades']:
                for key in got_profile.keys():
                    if upgrade[key]:
                        got_profile[key].append(upgrade['sdest'])

        self.maxDiff = None
        self.assertDictEqual(
            expected, got,
            'Unexpected gatherer result.\n\nPackages in result {0}:'.format(
                map(lambda profile: profile['id'], result)))

    def assert_outdated_profiles(self, expected_profiles, ignore=()):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles()
        got_profiles = [profile['id'] for profile in result
                        if profile['outdated_fs_version'] and profile['id'] not in ignore]
        self.assertEquals(expected_profiles, got_profiles,
                          'Unexpected outdated fs versions for profiles.')


class TestExtendAutoUpgradesWithHumanFormattedDateVersion(TestCase):

    def test_formats_timestamp_source_versions(self):
        input = {'ssource': '20141214183059',
                 'sdest': '1'}
        expected = {'ssource': '20141214183059',
                    'fsource': '2014/12/14 18:30',
                    'sdest': '1'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_formats_timestamp_dest_versions(self):
        input = {'ssource': '1',
                 'sdest': '20141214183059'}
        expected = {'ssource': '1',
                    'sdest': '20141214183059',
                    'fdest': '2014/12/14 18:30'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_does_not_format_non_timestamp_versions(self):
        input = {'ssource': '1',
                 'sdest': '2'}
        expected = {'ssource': '1',
                    'sdest': '2'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_does_not_fail_on_non_timestamp_versions_looking_like_timestamps(self):
        input = {'ssource': '10000000000000',
                 'sdest': '1'}
        expected = {'ssource': '10000000000000',
                    'sdest': '1'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

        input = {'ssource': '1',
                 'sdest': '10000000000000'}
        expected = {'ssource': '1',
                    'sdest': '10000000000000'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def format_upgrade_step(self, upgrade_step):
        profiles = [{'upgrades': [upgrade_step]}]
        output = extend_auto_upgrades_with_human_formatted_date_version(profiles)
        return output[0]['upgrades'][0]



from AccessControl import getSecurityManager
from ftw.upgrade.jsonapi.base import APIView
from ftw.upgrade.jsonapi.utils import action
from ftw.upgrade.jsonapi.utils import jsonify


class ZopeAppAPI(APIView):

    @jsonify
    @action('GET')
    def list_plone_sites(self):
        """Returns a list of Plone sites.
        """

        return list(self._get_plone_sites())

    @jsonify
    @action('GET')
    def current_user(self):
        """Return the current user when authenticated properly.
        This can be used for testing authentication.
        """
        return getSecurityManager().getUser().getId()

    def _get_plone_sites(self):
        overview_view = self.context.restrictedTraverse('plone-overview')
        for site in overview_view.sites():
            yield {'id': site.getId(),
                   'path': '/'.join(site.getPhysicalPath()),
                   'title': site.Title()}



from ftw.upgrade.interfaces import IUpgradeStepRecorder
from ftw.upgrade.tests.base import UpgradeTestCase
from zope.component import getMultiAdapter


class TestUpgradeStepRecorder(UpgradeTestCase):

    def test_marking_upgrades_as_installed(self):
        recorder = getMultiAdapter((self.layer['portal'], 'some.package:default'),
                                   IUpgradeStepRecorder)

        self.assertFalse(recorder.is_installed('20140101083000'))
        recorder.mark_as_installed('20140101083000')
        self.assertTrue(recorder.is_installed('20140101083000'))



from ftw.builder import Builder
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import statusmessages
from ftw.upgrade.browser.manage import ResponseLogger
from ftw.upgrade.tests.base import UpgradeTestCase
from plone.app.testing import SITE_OWNER_NAME
from Products.CMFCore.utils import getToolByName
from StringIO import StringIO
from unittest2 import TestCase
import logging
import re
import transaction


class TestResponseLogger(TestCase):

    def test_logging(self):
        response = StringIO()

        with ResponseLogger(response):
            logging.error('foo')
            logging.error('bar')

        response.seek(0)
        self.assertEqual(
            ['foo', u'bar'],
            response.read().strip().split('\n'))

    def test_logging_exceptions(self):
        response = StringIO()

        with self.assertRaises(KeyError):
            with ResponseLogger(response):
                raise KeyError('foo')

        response.seek(0)
        output = response.read().strip()
        # Dynamically replace paths so that it works on all machines
        output = re.sub(r'(File ").*(ftw/upgrade/.*")',
                        r'\1/.../\2', output)
        output = re.sub(r'(line )\d*', r'line XX', output)

        self.assertEqual(
            ['FAILED',
             'Traceback (most recent call last):',
             '  File "/.../ftw/upgrade/tests/'
             'test_manage_view.py", line XX, in test_logging_exceptions',
             "    raise KeyError('foo')",
             "KeyError: 'foo'"],
            output.split('\n'))

    def test_annotate_result_on_success(self):
        response = StringIO()

        with ResponseLogger(response, annotate_result=True):
            logging.error('foo')
            logging.error('bar')

        response.seek(0)
        self.assertEqual(
            ['foo', u'bar', 'Result: SUCCESS'],
            response.read().strip().split('\n'))

    def test_annotate_result_on_error(self):
        response = StringIO()

        with self.assertRaises(KeyError):
            with ResponseLogger(response, annotate_result=True):
                raise KeyError('foo')

        response.seek(0)
        output = response.read().strip()
        # Dynamically replace paths so that it works on all machines
        output = re.sub(r'(File ").*(ftw/upgrade/.*")',
                        r'\1/.../\2', output)
        output = re.sub(r'(line )\d*', r'line XX', output)

        self.assertEqual(
            ['FAILED',
             'Traceback (most recent call last):',
             '  File "/.../ftw/upgrade/tests/'
             'test_manage_view.py", line XX, in test_annotate_result_on_error',
             "    raise KeyError('foo')",
             "KeyError: 'foo'",
             'Result: FAILURE'],
            output.split('\n'))


class TestManageUpgrades(UpgradeTestCase):

    def setUp(self):
        super(TestManageUpgrades, self).setUp()
        self.portal_url = self.layer['portal'].portal_url()
        self.portal = self.layer['portal']

    @browsing
    def test_registered_in_controlpanel(self, browser):
        browser.login(SITE_OWNER_NAME).open(view='overview-controlpanel')
        link = browser.css('#content').find('Upgrades').first
        self.assertEqual(self.portal_url + '/@@manage-upgrades', link.attrib['href'])

    @browsing
    def test_manage_view_renders(self, browser):
        browser.login(SITE_OWNER_NAME).open(view='manage-upgrades')

        up_link = browser.css('#content').find('Up to Site Setup').first
        self.assertEqual(self.portal_url + '/@@overview-controlpanel',
                         up_link.attrib['href'])

        self.assertTrue(browser.css('input[value="plone.app.discussion:default"]').first)

    @browsing
    def test_manage_plain_view_renders(self, browser):
        browser.login(SITE_OWNER_NAME).open(view='manage-upgrades-plain')
        self.assertTrue(browser.css('input[value="plone.app.discussion:default"]').first)

    @browsing
    def test_install(self, browser):
        def upgrade_step(setup_context):
            portal = getToolByName(setup_context, 'portal_url').getPortalObject()
            portal.upgrade_installed = True

        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1111', '2222')
                                                .calling(upgrade_step, getToolByName)))

        with self.package_created():
            self.install_profile('the.package:default', '1111')
            self.portal.upgrade_installed = False
            transaction.commit()

            transaction.begin()  # sync transaction
            self.assertFalse(self.portal.upgrade_installed)

            browser.login(SITE_OWNER_NAME).open(view='manage-upgrades')
            # Install proposed upgrades
            browser.find('Install').click()

            transaction.begin()  # sync transaction
            self.assertTrue(self.portal.upgrade_installed)

    @browsing
    def test_upgrades_view_shows_cyclic_dependencies_error(self, browser):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('foo')
                                  .with_dependencies('the.package:bar'))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('bar')
                                  .with_dependencies('the.package:foo'))

        with self.package_created():
            self.install_profile('the.package:foo')
            self.install_profile('the.package:bar')
            transaction.commit()

            browser.login(SITE_OWNER_NAME).open(view='@@manage-upgrades')
            statusmessages.assert_message('There are cyclic dependencies.'
                                          ' The profiles could not be sorted'
                                          ' by dependencies!')

            possibilities = (
                ['the.package:foo ; the.package:bar'],
                ['the.package:bar ; the.package:foo'])

            self.assertIn(browser.css('.cyclic-dependencies li').text, possibilities)



from StringIO import StringIO
from path import Path
import contextlib
import os
import sys


def find_egginfo(path=None):
    path = path or Path(os.getcwd())
    if not path or path == '/':
        print >>sys.stderr, 'WARNING: no *.egg-info directory could be found.'
        return None

    egginfos = path.dirs('*.egg-info')
    if len(egginfos) == 0:
        return find_egginfo(path.dirname())

    if len(egginfos) > 1:
        print >>sys.stderr, 'WARNING: more than one *.egg-info' + \
            ' directory found.'
        return None

    return egginfos[0]


def find_package_namespace_path(egginfo):
    with egginfo.joinpath('top_level.txt').open() as top_level_file:
        top_level_path = top_level_file.read().strip()

    return egginfo.dirname().joinpath(top_level_path)


@contextlib.contextmanager
def capture():
    oldout = sys.stdout
    try:
        sys.stdout = StringIO()
        yield sys.stdout
    finally:
        sys.stdout = oldout



from ftw.builder import Builder
from ftw.builder.utils import serialize_callable
from ftw.upgrade import UpgradeStep
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName


class TestImportProfileUpgradeStepDirective(UpgradeTestCase):

    def setUp(self):
        super(TestImportProfileUpgradeStepDirective, self).setUp()
        self.portal_actions = getToolByName(self.portal, 'portal_actions')

    def test_upgrade_step_directive_without_handler(self):
        self.package.with_profile(Builder('genericsetup profile').with_fs_version('2'))
        self.package.with_zcml_include('ftw.upgrade', file='meta.zcml')
        self.package.with_zcml_node('upgrade-step:importProfile',
                                    title='Add test action.',
                                    profile='the.package:default',
                                    source='1',
                                    destination='2',
                                    directory='upgrade-profile/2')
        self.package.with_file('upgrade-profile/2/actions.xml',
                               self.asset('test-action.xml'), makedirs=True)

        with self.package_created():
            self.install_profile('the.package:default', '1')
            self.assertIsNone(self.get_action())
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('The Test Action', self.get_action().title)

    def test_upgrade_step_directive_with_handler(self):
        class Upgrade(UpgradeStep):
            def __call__(self):
                self.install_upgrade_profile()
                portal_actions = self.getToolByName('portal_actions')
                action = portal_actions.portal_tabs.get('test-action')
                assert action, 'The "test-action" was not created on import.'
                action.title = 'Title was changed.'

        self.package.with_profile(Builder('genericsetup profile').with_fs_version('2'))
        self.package.with_zcml_include('ftw.upgrade', file='meta.zcml')
        self.package.with_zcml_node('upgrade-step:importProfile',
                                    title='Add and rename test action.',
                                    profile='the.package:default',
                                    source='1',
                                    destination='2',
                                    directory='upgrade-profile/1',
                                    handler='.to2.Upgrade')
        self.package.with_file('to2.py', serialize_callable(Upgrade))
        self.package.with_file('upgrade-profile/1/actions.xml',
                               self.asset('test-action.xml'), makedirs=True)

        with self.package_created():
            self.install_profile('the.package:default', '1')
            self.assertIsNone(self.get_action())
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('Title was changed.', self.get_action().title)

    def get_action(self):
        return self.portal_actions.portal_tabs.get('test-action')



from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.factory import _DEFAULT_PROFILE
import transaction


class TestPloneUpgradeNeededCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestPloneUpgradeNeededCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('plone_upgrade_needed --help')

    def test_plone_upgrade_already_uptodate(self):
        exitcode, output = self.upgrade_script(
            'plone_upgrade_needed -s plone')
        self.assertEquals(0, exitcode)
        self.assertIn(u'false', output)

    def test_plone_upgrade_needed(self):
        setup = getToolByName(self.portal, 'portal_setup')
        setup.setLastVersionForProfile(_DEFAULT_PROFILE, '4')
        transaction.commit()

        exitcode, output = self.upgrade_script(
            'plone_upgrade_needed -s plone')
        self.assertEquals(0, exitcode)
        self.assertIn(u'true', output)



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testing.layer import COMPONENT_REGISTRY_ISOLATION
from ftw.testing.layer import ConsoleScriptLayer
from ftw.testing.layer import TEMP_DIRECTORY
from operator import itemgetter
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_ZSERVER
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2
from Products.CMFCore.utils import getToolByName
from Products.SiteAccess.VirtualHostMonster import manage_addVirtualHostMonster
from zope.configuration import xmlconfig
# This import is not used, but it registers a builder that we need:
import ftw.upgrade.tests.builders
import pkg_resources

ftw.upgrade.tests.builders  # pyflakes


COMMAND_LAYER = ConsoleScriptLayer('ftw.upgrade',
                                   bases=(BUILDER_LAYER, ),
                                   name='ftw.upgrade:command')


class UpgradeLayer(PloneSandboxLayer):
    defaultBases = (COMPONENT_REGISTRY_ISOLATION,
                    BUILDER_LAYER,
                    TEMP_DIRECTORY)

    def setUpZope(self, app, configurationContext):
        import Products.CMFPlacefulWorkflow
        xmlconfig.file('configure.zcml', Products.CMFPlacefulWorkflow,
                       context=configurationContext)

        import ftw.upgrade
        xmlconfig.file('configure.zcml', ftw.upgrade,
                       context=configurationContext)

        z2.installProduct(app, 'Products.CMFPlacefulWorkflow')

        manage_addVirtualHostMonster(app, 'virtual_hosting')

    def setUpPloneSite(self, portal):
        applyProfile(
            portal, 'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow')
        applyProfile(portal, 'ftw.upgrade:default')

        self.fix_plone_app_jquery_version(portal)

    def fix_plone_app_jquery_version(self, portal):
        try:
            pkg_resources.get_distribution('plone.app.jquery')
        except pkg_resources.DistributionNotFound:
            return

        # The plone.app.jquery version shipped with Plone 4.2 has an outdated
        # metadata.xml version, resulting in proposed upgrades in a fresh
        # installation.
        # For consistent test result we fix that here.
        portal_setup = getToolByName(portal, 'portal_setup')
        profileid = 'plone.app.jquery:default'
        version = max(map(itemgetter('dest'),
                          portal_setup.listUpgrades(profileid, show_old=True)))
        portal_setup.setLastVersionForProfile(profileid, version)


UPGRADE_LAYER = UpgradeLayer()
UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.upgrade:functional")

COMMAND_AND_UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_ZSERVER,
           UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory),
           COMMAND_LAYER),
    name="ftw.upgrade:command_and_functional")


class IntIdUpgradeLayer(PloneSandboxLayer):

    defaultBases = (UPGRADE_LAYER, )

    def setUpZope(self, app, configurationContext):
        import plone.app.intid
        xmlconfig.file('configure.zcml', plone.app.intid,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.intid:default')


INTID_UPGRADE_LAYER = IntIdUpgradeLayer()
INTID_UPGRADE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(INTID_UPGRADE_LAYER,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.upgrade-intid:functional")



from ftw.upgrade import UpgradeStep
from ftw.upgrade.exceptions import UpgradeStepDefinitionError
from ftw.upgrade.utils import subject_from_docstring
from glob import glob
from Products.GenericSetup.upgrade import normalize_version
import imp
import inspect
import os.path
import re


UPGRADESTEP_DATETIME_REGEX = re.compile(r'^.*/?(\d{14})[^/]*/upgrade.py$')


class Scanner(object):

    def __init__(self, dottedname, directory):
        self.dottedname = dottedname
        self.directory = directory

    def scan(self):
        self._load_upgrades_directory()
        infos = map(self._build_upgrade_step_info,
                    self._find_upgrade_directories())
        infos.sort(key=lambda info: normalize_version(info['target-version']))
        if len(infos) > 0:
            reduce(self._chain_upgrade_steps, infos)
        return infos

    def _find_upgrade_directories(self):
        return filter(UPGRADESTEP_DATETIME_REGEX.match,
                      glob('{0}/*/upgrade.py'.format(self.directory)))

    def _build_upgrade_step_info(self, path):
        title, callable = self._load_upgrade_step_code(path)
        return {'source-version': None,
                'target-version':
                    UPGRADESTEP_DATETIME_REGEX.match(path).group(1),
                'path': os.path.dirname(path),
                'title': title,
                'callable': callable}

    def _chain_upgrade_steps(self, first, second):
        second['source-version'] = first['target-version']
        return second

    def _load_upgrade_step_code(self, upgrade_path):
        path = os.path.dirname(upgrade_path)

        try:
            fp, pathname, description = imp.find_module('.', [path])
        except ImportError:
            pass
        else:
            name = '.'.join((self.dottedname, os.path.basename(path)))
            imp.load_module(name, fp, pathname, description)

        fp, pathname, description = imp.find_module('upgrade', [path])
        name = '.'.join((self.dottedname,
                         os.path.basename(path),
                         'upgrade'))

        module = imp.load_module(name, fp, pathname, description)
        upgrade_steps = tuple(self._find_upgrade_step_classes_in_module(
                module))

        if len(upgrade_steps) == 0:
            raise UpgradeStepDefinitionError(
                'The upgrade step {0} has no upgrade class in the'
                ' upgrade.py module.'.format(os.path.basename(path)))

        if len(upgrade_steps) > 1:
            raise UpgradeStepDefinitionError(
                'The upgrade step {0} has more than one upgrade class in the'
                ' upgrade.py module.'.format(os.path.basename(path)))

        return upgrade_steps[0]

    def _find_upgrade_step_classes_in_module(self, module):
        for name, value in inspect.getmembers(module, inspect.isclass):
            if not issubclass(value, UpgradeStep):
                continue

            if inspect.getmodule(value) is not module:
                continue

            title = subject_from_docstring(inspect.getdoc(value) or name)
            if isinstance(title, str):
                title = title.decode('utf-8')
            yield (title, value)

    def _load_upgrades_directory(self):
        """This method tries to load the upgrade step directory, if there is
        an __init__.py. This helps to avoid RuntimeWarnings.
        However, it is not relevant for anything to work,
        it just makes Python happy.
        """
        try:
            fp, pathname, description = imp.find_module('.', [self.directory])
        except ImportError:
            pass
        else:
            imp.load_module(self.dottedname, fp, pathname, description)



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade import UpgradeStep
from ftw.upgrade.interfaces import IExecutioner
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter


class TestDirectoryUpgradeSteps(UpgradeTestCase):

    def setUp(self):
        super(TestDirectoryUpgradeSteps, self).setUp()
        self.portal_actions = getToolByName(self.portal, 'portal_actions')

    def test_installing_profile_sets_version_to_latest_upgrade(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2014, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default')
            version = self.portal_setup.getLastVersionForProfile('the.package:default')
            self.assertEqual(('20140101000000',), version)

    def test_installing_profile_marks_all_upgrade_steps_as_installed_too(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            self.install_profile('the.package:default')

            self.assertTrue(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to be marked as installed'
                ' after importing profile.')

    def test_associated_upgrade_step_profile_is_imported(self):
        class AddTestAction(UpgradeStep):
            def __call__(self):
                self.install_upgrade_profile()

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .with_file('actions.xml', self.asset('test-action.xml'))
                          .calling(AddTestAction)))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertIsNone(self.get_action())
            self.install_profile_upgrades('the.package:default')
            self.assertEqual('The Test Action', self.get_action().title)

    def test_running_upgrade_steps_marks_them_as_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('ftw upgrade step')
                                                .to(datetime(2011, 1, 1))))

        with self.package_created():
            recorder = getMultiAdapter((self.portal, 'the.package:default'),
                                       IUpgradeStepRecorder)

            self.portal_setup.setLastVersionForProfile('the.package:default', u'0')
            self.assertFalse(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to not be marked as installed'
                ' before importing profile.')

            executioner = IExecutioner(self.portal_setup)
            first_upgrade, = self.portal_setup.listUpgrades('the.package:default')
            executioner.install([('the.package:default', [first_upgrade['id']])])

            self.assertTrue(
                recorder.is_installed('20110101000000'),
                'Expected upgrade steps to be marked as installed'
                ' after upgrading.')

    def test_base_profile_and_target_version_are_provided(self):
        self.portal.upgrade_infos = {}

        class StoreUpgradeInfos(UpgradeStep):
            def __call__(self):
                self.portal.upgrade_infos.update({
                    'base_profile': self.base_profile,
                    'target_version': self.target_version,
                })

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .calling(StoreUpgradeInfos)))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertEquals({}, self.portal.upgrade_infos)
            self.install_profile_upgrades('the.package:default')
            self.assertEquals(
                {'base_profile': u'profile-the.package:default',
                 'target_version': '20110101000000'},
                self.portal.upgrade_infos)

    def get_action(self):
        return self.portal_actions.portal_tabs.get('test-action')
